"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test DiscRates class.
"""
import unittest
import numpy as np

from climada import CONFIG
from climada.entity.disc_rates.base import DiscRates
from climada.util.constants import ENT_TEMPLATE_XLS, ENT_DEMO_TODAY

ENT_TEST_MAT = CONFIG.exposures.test_data.dir().joinpath('demo_today.mat')

class TestChecker(unittest.TestCase):
    """Test discount rates attributes checker"""

    def test_check_wrongRates_fail(self):
        """Wrong discount rates definition"""
        disc_rate = DiscRates()
        disc_rate.rates = np.array([3, 4])
        disc_rate.years = np.array([1])

        with self.assertRaises(ValueError) as cm:
            disc_rate.check()
        self.assertIn('Invalid DiscRates.rates size: 1 != 2.', str(cm.exception))

class TestConstructor(unittest.TestCase):
    """Test discount rates attributes."""
    def test_attributes_all(self):
        """All attributes are defined"""
        disc_rate = DiscRates()
        self.assertTrue(hasattr(disc_rate, 'years'))
        self.assertTrue(hasattr(disc_rate, 'rates'))

class TestAppend(unittest.TestCase):
    """Check append function"""
    def test_append_to_empty_same(self):
        """Append DiscRates to empty one."""
        disc_rate = DiscRates()
        disc_rate_add = DiscRates()
        disc_rate_add.tag.file_name = 'file1.txt'
        disc_rate_add.tag.description = 'descr1'
        disc_rate_add.years = np.array([2000, 2001, 2002])
        disc_rate_add.rates = np.array([0.1, 0.2, 0.3])

        disc_rate.append(disc_rate_add)
        disc_rate.check()

        self.assertTrue(np.array_equal(disc_rate.years, disc_rate_add.years))
        self.assertTrue(np.array_equal(disc_rate.rates, disc_rate_add.rates))
        self.assertTrue(np.array_equal(disc_rate.tag.file_name,
                                       disc_rate_add.tag.file_name))
        self.assertTrue(np.array_equal(disc_rate.tag.description,
                                       disc_rate_add.tag.description))

    def test_append_equal_same(self):
        """Append the same DiscRates. The inital DiscRates is obtained."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.array([2000, 2001, 2002])
        disc_rate.rates = np.array([0.1, 0.2, 0.3])

        disc_rate_add = DiscRates()
        disc_rate_add.tag.file_name = 'file1.txt'
        disc_rate_add.tag.description = 'descr1'
        disc_rate_add.years = np.array([2000, 2001, 2002])
        disc_rate_add.rates = np.array([0.1, 0.2, 0.3])

        disc_rate.append(disc_rate_add)
        disc_rate.check()

        self.assertTrue(np.array_equal(disc_rate.years, disc_rate_add.years))
        self.assertTrue(np.array_equal(disc_rate.rates, disc_rate_add.rates))
        self.assertTrue(np.array_equal(disc_rate.tag.file_name, disc_rate_add.tag.file_name))
        self.assertEqual(disc_rate.tag.description, disc_rate_add.tag.description)

    def test_append_different_append(self):
        """Append DiscRates with same and new values. The rates with repeated
        years are overwritten."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.array([2000, 2001, 2002])
        disc_rate.rates = np.array([0.1, 0.2, 0.3])

        disc_rate_add = DiscRates()
        disc_rate_add.tag.file_name = 'file2.txt'
        disc_rate_add.tag.description = 'descr2'
        disc_rate_add.years = np.array([2000, 2001, 2003])
        disc_rate_add.rates = np.array([0.11, 0.22, 0.33])

        disc_rate.append(disc_rate_add)
        disc_rate.check()

        self.assertTrue(np.array_equal(disc_rate.years,
                                       np.array([2000, 2001, 2002, 2003])))
        self.assertTrue(np.array_equal(disc_rate.rates,
                                       np.array([0.11, 0.22, 0.3, 0.33])))
        self.assertTrue(np.array_equal(disc_rate.tag.file_name, 'file1.txt + file2.txt'))
        self.assertTrue(np.array_equal(disc_rate.tag.description, 'descr1 + descr2'))

class TestSelect(unittest.TestCase):
    """Test select method"""
    def test_select_pass(self):
        """Test select right time range."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.arange(2000, 2050)
        disc_rate.rates = np.arange(disc_rate.years.size)

        year_range = np.arange(2010, 2020)
        sel_disc = disc_rate.select(year_range)

        self.assertTrue(np.array_equal(sel_disc.years, year_range))
        self.assertTrue(np.array_equal(sel_disc.rates, disc_rate.rates[10:20]))

    def test_select_wrong_pass(self):
        """Test select wrong time range."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.arange(2000, 2050)
        disc_rate.rates = np.arange(disc_rate.years.size)

        year_range = np.arange(2050, 2060)
        self.assertEqual(None, disc_rate.select(year_range))

class TestNetPresValue(unittest.TestCase):
    """Test select method"""
    def test_net_present_value_pass(self):
        """Test net_present_value right time range."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.arange(2000, 2050)
        disc_rate.rates = np.ones(disc_rate.years.size) * 0.02

        val_years = np.ones(23) * 6.512201157564418e9
        res = disc_rate.net_present_value(2018, 2040, val_years)
        self.assertEqual(res, 1.215049630691397e+11)

    def test_net_present_value_wrong_pass(self):
        """Test net_present_value wrong time range."""
        disc_rate = DiscRates()
        disc_rate.tag.file_name = 'file1.txt'
        disc_rate.tag.description = 'descr1'
        disc_rate.years = np.arange(2000, 2050)
        disc_rate.rates = np.arange(disc_rate.years.size)
        val_years = np.ones(11) * 6.512201157564418e9
        with self.assertRaises(ValueError):
            disc_rate.net_present_value(2050, 2060, val_years)

class TestReaderExcel(unittest.TestCase):
    """Test excel reader for discount rates"""

    def test_demo_file_pass(self):
        """Read demo excel file."""
        description = 'One single file.'
        disc_rate = DiscRates.from_excel(ENT_DEMO_TODAY, description)

        # Check results
        n_rates = 51

        self.assertIn('int', str(disc_rate.years.dtype))
        self.assertEqual(disc_rate.years.shape, (n_rates,))
        self.assertEqual(disc_rate.years[0], 2000)
        self.assertEqual(disc_rate.years[n_rates - 1], 2050)

        self.assertIn('float', str(disc_rate.rates.dtype))
        self.assertEqual(disc_rate.rates.shape, (n_rates,))
        self.assertEqual(disc_rate.rates.min(), 0.02)
        self.assertEqual(disc_rate.rates.max(), 0.02)

        self.assertEqual(disc_rate.tag.file_name, str(ENT_DEMO_TODAY))
        self.assertEqual(disc_rate.tag.description, description)

    def test_template_file_pass(self):
        """Read demo excel file."""
        disc_rate = DiscRates.from_excel(ENT_TEMPLATE_XLS)

        # Check results
        n_rates = 102

        self.assertIn('int', str(disc_rate.years.dtype))
        self.assertEqual(disc_rate.years.shape, (n_rates,))
        self.assertEqual(disc_rate.years[0], 2000)
        self.assertEqual(disc_rate.years[n_rates - 1], 2101)

        self.assertIn('float', str(disc_rate.rates.dtype))
        self.assertEqual(disc_rate.rates.shape, (n_rates,))
        self.assertEqual(disc_rate.rates.min(), 0.02)
        self.assertEqual(disc_rate.rates.max(), 0.02)

        self.assertEqual(disc_rate.tag.file_name, str(ENT_TEMPLATE_XLS))
        self.assertEqual(disc_rate.tag.description, '')

class TestReaderMat(unittest.TestCase):
    """Test mat reader for discount rates"""

    def test_demo_file_pass(self):
        """Read demo mat file"""
        # Read demo excel file

        description = 'One single file.'
        disc_rate = DiscRates.from_mat(file_name=ENT_TEST_MAT, description=description)

        # Check results
        n_rates = 51

        self.assertIn('int', str(disc_rate.years.dtype))
        self.assertEqual(len(disc_rate.years), n_rates)
        self.assertEqual(disc_rate.years[0], 2000)
        self.assertEqual(disc_rate.years[n_rates - 1], 2050)

        self.assertIn('float', str(disc_rate.rates.dtype))
        self.assertEqual(len(disc_rate.rates), n_rates)
        self.assertEqual(disc_rate.rates.min(), 0.02)
        self.assertEqual(disc_rate.rates.max(), 0.02)

        self.assertEqual(disc_rate.tag.file_name, str(ENT_TEST_MAT))
        self.assertEqual(disc_rate.tag.description, description)


class TestWriter(unittest.TestCase):
    """Test excel reader for discount rates"""

    def test_write_read_pass(self):
        """Read demo excel file."""
        years = np.arange(1950, 2150)
        rates = np.ones(years.size) * 0.03
        disc_rate = DiscRates(years=years, rates=rates)

        file_name = CONFIG.disc_rates.test_data.dir().joinpath('test_disc.xlsx')
        disc_rate.write_excel(file_name)

        disc_read = DiscRates.from_excel(file_name)

        self.assertTrue(np.array_equal(disc_read.years, disc_rate.years))
        self.assertTrue(np.array_equal(disc_read.rates, disc_rate.rates))

        self.assertEqual(disc_read.tag.file_name, str(file_name))
        self.assertEqual(disc_read.tag.description, '')

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestChecker)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestAppend))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestConstructor))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSelect))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestNetPresValue))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderExcel))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderMat))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWriter))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test save module.
"""
import unittest
from pathlib import Path

from climada import CONFIG
from climada.util.save import save, load

DATA_DIR = CONFIG.util.test_data.str()
IN_CONFIG = CONFIG.local_data.save_dir.str()

class TestSave(unittest.TestCase):
    """Test save function"""

    def setUp(self):
        CONFIG.local_data.save_dir._val = DATA_DIR

    def tearDown(self):
        CONFIG.local_data.save_dir._val = IN_CONFIG

    def test_entity_in_save_dir(self):
        """Returns the same list if its length is correct."""
        file_name = 'save_test.pkl'
        ent = {'value': [1, 2, 3]}
        with self.assertLogs('climada.util.save', level='INFO') as cm:
            save(file_name, ent)
        self.assertTrue(CONFIG.local_data.save_dir.dir().joinpath(file_name).is_file())
        self.assertTrue((file_name in cm.output[0]) or
                        (file_name in cm.output[1]))

    def test_load_pass(self):
        """Load previously saved variable"""
        file_name = 'save_test.pkl'
        ent = {'value': [1, 2, 3]}
        save(file_name, ent)
        res = load(file_name)
        self.assertTrue('value' in res)
        self.assertTrue(res['value'] == ent['value'])

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestSave)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test CostBenefit class.
"""
from pathlib import Path
import copy
import unittest
import numpy as np

from climada.entity.entity_def import Entity
from climada.entity.disc_rates import DiscRates
from climada.hazard.base import Hazard
from climada.engine.cost_benefit import CostBenefit, risk_aai_agg, \
        risk_rp_100, risk_rp_250, _norm_values
from climada.engine import Impact
from climada.util.constants import ENT_DEMO_FUTURE, ENT_DEMO_TODAY
from climada.util.api_client import Client


def get_test_file(ds_name):
    client = Client()
    test_ds = client.get_dataset_info(name=ds_name, status='test_dataset')
    _, test_files = client.download_dataset(test_ds)
    return test_files[0]


HAZ_TEST_MAT = get_test_file('atl_prob_no_name')
ENT_TEST_MAT = get_test_file('demo_today')


class TestSteps(unittest.TestCase):
    """Test intermediate steps"""
    def test_calc_impact_measures_pass(self):
        """Test _calc_impact_measures against reference value"""
        self.assertTrue(HAZ_TEST_MAT.is_file(), "{} is not a file".format(HAZ_TEST_MAT))
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        self.assertTrue(ENT_TEST_MAT.is_file(), "{} is not a file".format(ENT_TEST_MAT))
        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.check()
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        entity.check()

        cost_ben = CostBenefit()
        cost_ben._calc_impact_measures(hazard, entity.exposures, entity.measures,
                                       entity.impact_funcs, when='future',
                                       risk_func=risk_aai_agg, save_imp=True)

        self.assertEqual(cost_ben.imp_meas_present, dict())
        self.assertEqual(cost_ben.cost_ben_ratio, dict())
        self.assertEqual(cost_ben.benefit, dict())
        self.assertEqual(cost_ben.tot_climate_risk, 0.0)
        self.assertEqual(cost_ben.present_year, 2016)
        self.assertEqual(cost_ben.future_year, 2030)

        self.assertEqual(cost_ben.imp_meas_future['no measure']['cost'], (0, 0))
        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['risk'],
                               6.51220115756442e+09, places=3)
        new_efc = cost_ben.imp_meas_future['no measure']['impact'].calc_freq_curve()
        self.assertTrue(
            np.allclose(new_efc.return_per,
                        cost_ben.imp_meas_future['no measure']['efc'].return_per))
        self.assertTrue(
            np.allclose(new_efc.impact, cost_ben.imp_meas_future['no measure']['efc'].impact))
        self.assertEqual(
            cost_ben.imp_meas_future['no measure']['impact'].at_event.nonzero()[0].size,
            841)
        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['impact'].at_event[14082],
                               8.801682862431524e+06, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['impact'].tot_value,
                               6.570532945599105e+11, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['impact'].aai_agg,
                               6.51220115756442e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['cost'][0],
                               1.3117683608515418e+09, places=3)
        self.assertEqual(cost_ben.imp_meas_future['Mangroves']['cost'][1], 1)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['risk'],
                               4.850407096284983e+09, places=3)
        new_efc = cost_ben.imp_meas_future['Mangroves']['impact'].calc_freq_curve()
        self.assertTrue(
            np.allclose(new_efc.return_per,
                        cost_ben.imp_meas_future['Mangroves']['efc'].return_per))
        self.assertTrue(
            np.allclose(new_efc.impact, cost_ben.imp_meas_future['Mangroves']['efc'].impact))
        self.assertEqual(
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event.nonzero()[0].size,
            665)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['impact'].at_event[13901],
                               1.29576562770977e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['impact'].tot_value,
                               6.570532945599105e+11, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['impact'].aai_agg,
                               4.850407096284983e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['cost'][0],
                               1.728000000000000e+09, places=3)
        self.assertEqual(cost_ben.imp_meas_future['Beach nourishment']['cost'][1], 1)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['risk'],
                               5.188921355413834e+09, places=3)
        new_efc = cost_ben.imp_meas_future['Beach nourishment']['impact'].calc_freq_curve()
        self.assertTrue(
            np.allclose(new_efc.return_per,
                        cost_ben.imp_meas_future['Beach nourishment']['efc'].return_per))
        self.assertTrue(
            np.allclose(new_efc.impact,
                        cost_ben.imp_meas_future['Beach nourishment']['efc'].impact))
        self.assertEqual(
            cost_ben.imp_meas_future['Beach nourishment']['impact'].at_event.nonzero()[0].size,
            702)
        self.assertEqual(cost_ben.imp_meas_future['Beach nourishment']['impact'].at_event[1110],
                         0.0)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['impact'].eai_exp[5],
                               1.1133679079730146e+08, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['impact'].tot_value,
                               6.570532945599105e+11, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['impact'].aai_agg,
                               5.188921355413834e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['cost'][0],
                               8.878779433630093e+09, places=3)
        self.assertEqual(cost_ben.imp_meas_future['Seawall']['cost'][1], 1)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['risk'],
                               4.736400526119911e+09, places=3)
        new_efc = cost_ben.imp_meas_future['Seawall']['impact'].calc_freq_curve()
        self.assertTrue(np.allclose(new_efc.return_per,
                                    cost_ben.imp_meas_future['Seawall']['efc'].return_per))
        self.assertTrue(np.allclose(new_efc.impact,
                                    cost_ben.imp_meas_future['Seawall']['efc'].impact))
        self.assertEqual(cost_ben.imp_meas_future['Seawall']['impact'].at_event.nonzero()[0].size,
                         73)
        self.assertEqual(cost_ben.imp_meas_future['Seawall']['impact'].at_event[1229], 0.0)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['impact'].tot_value,
                               6.570532945599105e+11, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['impact'].aai_agg,
                               4.736400526119911e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['cost'][0],
                               9.200000000000000e+09, places=3)
        self.assertEqual(cost_ben.imp_meas_future['Building code']['cost'][1], 1)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['risk'],
                               4.884150868173321e+09, places=3)
        new_efc = cost_ben.imp_meas_future['Building code']['impact'].calc_freq_curve()
        self.assertTrue(np.allclose(new_efc.return_per,
                                    cost_ben.imp_meas_future['Building code']['efc'].return_per))
        self.assertTrue(np.allclose(new_efc.impact,
                                    cost_ben.imp_meas_future['Building code']['efc'].impact))
        self.assertEqual(
            cost_ben.imp_meas_future['Building code']['impact'].at_event.nonzero()[0].size,
            841)
        self.assertEqual(cost_ben.imp_meas_future['Building code']['impact'].at_event[122], 0.0)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['impact'].eai_exp[11],
                               7.757060129393841e+07, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['impact'].tot_value,
                               6.570532945599105e+11, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['impact'].aai_agg,
                               4.884150868173321e+09, places=3)

    def test_cb_one_meas_pres_pass(self):
        """Test _cost_ben_one with different future"""
        meas_name = 'Mangroves'
        meas_val = dict()
        meas_val['cost'] = (1.3117683608515418e+09, 1)
        meas_val['risk'] = 4.826231151473135e+10
        meas_val['efc'] = None
        meas_val['risk_transf'] = 0

        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2040
        cb.imp_meas_present['no measure'] = dict()
        cb.imp_meas_present['no measure']['risk'] = 6.51220115756442e+09
        cb.imp_meas_present['Mangroves'] = dict()
        cb.imp_meas_present['Mangroves']['risk'] = 4.850407096284983e+09
        cb.imp_meas_present['Mangroves']['risk_transf'] = 0

        cb.imp_meas_future['no measure'] = dict()
        cb.imp_meas_future['no measure']['risk'] = 5.9506659786664024e+10

        disc_rates = DiscRates()
        disc_rates.years = np.arange(2016, 2051)
        disc_rates.rates = np.ones(disc_rates.years.size) * 0.02

        time_dep = cb._time_dependency_array(1)

        cb._cost_ben_one(meas_name, meas_val, disc_rates, time_dep)
        self.assertAlmostEqual(cb.benefit[meas_name], 113345027690.81276, places=3)
        self.assertAlmostEqual(cb.cost_ben_ratio[meas_name], 0.011573232523528404)

    def test_cb_one_meas_fut_pass(self):
        """Test _cost_ben_one with same future"""
        meas_name = 'Mangroves'
        meas_val = dict()
        meas_val['cost'] = (1.3117683608515418e+09, 1)
        meas_val['risk'] = 4.850407096284983e+09
        meas_val['efc'] = None
        meas_val['risk_transf'] = 0

        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2040
        cb.imp_meas_future['no measure'] = dict()
        cb.imp_meas_future['no measure']['risk'] = 6.51220115756442e+09

        disc_rates = DiscRates()
        disc_rates.years = np.arange(2000, 2051)
        disc_rates.rates = np.ones(disc_rates.years.size) * 0.02

        time_dep = cb._time_dependency_array()

        cb._cost_ben_one(meas_name, meas_val, disc_rates, time_dep)
        self.assertAlmostEqual(cb.benefit[meas_name], 3.100583368954022e+10, places=3)
        self.assertAlmostEqual(cb.cost_ben_ratio[meas_name], 0.04230714690616641)

    def test_calc_cb_no_change_pass(self):
        """Test _calc_cost_benefit without present value against reference value"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        entity.check()

        cost_ben = CostBenefit()
        cost_ben._calc_impact_measures(hazard, entity.exposures, entity.measures,
                                       entity.impact_funcs, when='future',
                                       risk_func=risk_aai_agg, save_imp=True)

        cost_ben.present_year = 2018
        cost_ben.future_year = 2040
        cost_ben._calc_cost_benefit(entity.disc_rates)

        self.assertEqual(cost_ben.imp_meas_present, dict())
        self.assertEqual(len(cost_ben.imp_meas_future), 5)
        self.assertEqual(cost_ben.present_year, 2018)
        self.assertEqual(cost_ben.future_year, 2040)

        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Mangroves'], 0.04230714690616641)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Beach nourishment'], 0.06998836431681373)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Seawall'], 0.2679741183248266)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Building code'], 0.30286828677985717)

        self.assertAlmostEqual(cost_ben.benefit['Mangroves'], 3.100583368954022e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Beach nourishment'],
                               2.468981832719974e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Seawall'], 3.3132973770502796e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Building code'], 3.0376240767284798e+10, places=3)

        self.assertAlmostEqual(cost_ben.tot_climate_risk, 1.2150496306913972e+11, places=3)

    def test_calc_cb_change_pass(self):
        """Test _calc_cost_benefit with present value against reference value"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        entity.check()

        cost_ben = CostBenefit()
        cost_ben._calc_impact_measures(hazard, entity.exposures, entity.measures,
                                       entity.impact_funcs, when='present',
                                       risk_func=risk_aai_agg, save_imp=False)

        ent_future = Entity.from_excel(ENT_DEMO_FUTURE)
        ent_future.check()

        haz_future = copy.deepcopy(hazard)
        haz_future.intensity.data += 25

        cost_ben._calc_impact_measures(haz_future, ent_future.exposures, ent_future.measures,
                                       ent_future.impact_funcs, when='future',
                                       risk_func=risk_aai_agg, save_imp=False)

        cost_ben.present_year = 2018
        cost_ben.future_year = 2040
        cost_ben._calc_cost_benefit(entity.disc_rates, imp_time_depen=1)

        self.assertEqual(cost_ben.present_year, 2018)
        self.assertEqual(cost_ben.future_year, 2040)
        self.assertAlmostEqual(cost_ben.tot_climate_risk, 5.768659152882021e+11, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_present['no measure']['risk'],
                               6.51220115756442e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Mangroves']['risk'],
                               4.850407096284983e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Beach nourishment']['risk'],
                               5.188921355413834e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Seawall']['risk'],
                               4.736400526119911e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Building code']['risk'],
                               4.884150868173321e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['risk'],
                               5.9506659786664024e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['risk'],
                               4.826231151473135e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['risk'],
                               5.0647250923231674e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['risk'],
                               21089567135.7345, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['risk'],
                               4.462999483999791e+10, places=3)

        self.assertAlmostEqual(cost_ben.benefit['Mangroves'], 113345027690.81276, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Beach nourishment'], 89444869971.53653, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Seawall'], 347977469896.1333, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Building code'], 144216478822.05154, places=3)

        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Mangroves'], 0.011573232523528404)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Beach nourishment'], 0.01931916274851638)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Seawall'], 0.025515385913577368)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Building code'], 0.06379298728650741)

        self.assertAlmostEqual(cost_ben.tot_climate_risk, 576865915288.2021, places=3)

    def test_time_array_pres_pass(self):
        """Test _time_dependency_array"""
        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2030
        imp_time_depen = 1.0
        time_arr = cb._time_dependency_array(imp_time_depen)

        n_years = cb.future_year - cb.present_year + 1
        self.assertEqual(time_arr.size, n_years)
        self.assertTrue(np.allclose(time_arr[:-1], np.arange(0, 1, 1 / (n_years - 1))))
        self.assertEqual(time_arr[-1], 1)

        imp_time_depen = 0.5
        time_arr = cb._time_dependency_array(imp_time_depen)

        n_years = cb.future_year - cb.present_year + 1
        self.assertEqual(time_arr.size, n_years)
        self.assertTrue(np.allclose(time_arr, np.arange(n_years)**imp_time_depen /
                                    (n_years - 1)**imp_time_depen))

    def test_time_array_no_pres_pass(self):
        """Test _time_dependency_array"""
        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2030
        time_arr = cb._time_dependency_array()

        n_years = cb.future_year - cb.present_year + 1
        self.assertEqual(time_arr.size, n_years)
        self.assertTrue(np.array_equal(time_arr, np.ones(n_years)))

    def test_npv_unaverted_no_pres_pass(self):
        """Test _npv_unaverted_impact"""
        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2030
        risk_future = 1000
        disc_rates = DiscRates()
        disc_rates.years = np.arange(cb.present_year, cb.future_year + 1)
        disc_rates.rates = np.ones(disc_rates.years.size) * 0.025
        time_dep = np.linspace(0, 1, disc_rates.years.size)
        res = cb._npv_unaverted_impact(risk_future, disc_rates, time_dep,
                                       risk_present=None)

        self.assertEqual(
            res,
            disc_rates.net_present_value(cb.present_year, cb.future_year, time_dep * risk_future))

    def test_npv_unaverted_pres_pass(self):
        """Test _npv_unaverted_impact"""
        cb = CostBenefit()
        cb.present_year = 2018
        cb.future_year = 2030
        risk_future = 1000
        risk_present = 500
        disc_rates = DiscRates()
        disc_rates.years = np.arange(cb.present_year, cb.future_year + 1)
        disc_rates.rates = np.ones(disc_rates.years.size) * 0.025
        time_dep = np.linspace(0, 1, disc_rates.years.size)
        res = cb._npv_unaverted_impact(risk_future, disc_rates, time_dep, risk_present)


        tot_climate_risk = risk_present + (risk_future - risk_present) * time_dep
        self.assertEqual(res, disc_rates.net_present_value(cb.present_year,
                                                           cb.future_year,
                                                           tot_climate_risk))

    def test_norm_value(self):
        """Test _norm_values"""
        norm_fact, norm_name = _norm_values(1)
        self.assertEqual(norm_fact, 1)
        self.assertEqual(norm_name, "")

        norm_fact, norm_name = _norm_values(10)
        self.assertEqual(norm_fact, 1)
        self.assertEqual(norm_name, "")

        norm_fact, norm_name = _norm_values(100)
        self.assertEqual(norm_fact, 1)
        self.assertEqual(norm_name, "")

        norm_fact, norm_name = _norm_values(1001)
        self.assertEqual(norm_fact, 1000)
        self.assertEqual(norm_name, "k")

        norm_fact, norm_name = _norm_values(10000)
        self.assertEqual(norm_fact, 1000)
        self.assertEqual(norm_name, "k")

        norm_fact, norm_name = _norm_values(1.01e6)
        self.assertEqual(norm_fact, 1.0e6)
        self.assertEqual(norm_name, "m")

        norm_fact, norm_name = _norm_values(1.0e8)
        self.assertEqual(norm_fact, 1.0e6)
        self.assertEqual(norm_name, "m")

        norm_fact, norm_name = _norm_values(1.01e9)
        self.assertEqual(norm_fact, 1.0e9)
        self.assertEqual(norm_name, "bn")

        norm_fact, norm_name = _norm_values(1.0e10)
        self.assertEqual(norm_fact, 1.0e9)
        self.assertEqual(norm_name, "bn")

        norm_fact, norm_name = _norm_values(1.0e12)
        self.assertEqual(norm_fact, 1.0e9)
        self.assertEqual(norm_name, "bn")

    def test_combine_fut_pass(self):
        """Test combine_measures with present and future"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018

        fut_ent = copy.deepcopy(entity)
        fut_ent.exposures.ref_year = 2040
        fut_haz = copy.deepcopy(hazard)

        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, fut_haz, fut_ent, future_year=None,
                      risk_func=risk_aai_agg, imp_time_depen=None, save_imp=True)

        new_name = 'combine'
        new_color = np.array([0.1, 0.1, 0.1])
        new_cb = cost_ben.combine_measures(['Mangroves', 'Seawall'], new_name, new_color,
                                           entity.disc_rates, imp_time_depen=None,
                                           risk_func=risk_aai_agg)

        self.assertTrue(np.allclose(new_cb.color_rgb[new_name], new_color))

        new_imp = cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event
        new_imp += cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Seawall']['impact'].at_event
        new_imp = np.maximum(cost_ben.imp_meas_future['no measure']['impact'].at_event - new_imp,
                             0)

        self.assertTrue(np.allclose(new_cb.imp_meas_present[new_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_present[new_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_present['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(new_cb.imp_meas_present[new_name]['cost'][0],
                               cost_ben.imp_meas_present['Mangroves']['cost'][0] +
                               cost_ben.imp_meas_present['Seawall']['cost'][0])
        self.assertAlmostEqual(new_cb.imp_meas_present[new_name]['cost'][1], 1)
        self.assertTrue(np.allclose(
            new_cb.imp_meas_present[new_name]['efc'].impact,
            new_cb.imp_meas_present[new_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_present[new_name]['risk_transf'], 0)

        self.assertTrue(np.allclose(new_cb.imp_meas_future[new_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_future[new_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(
            new_cb.imp_meas_future[new_name]['cost'][0],
            cost_ben.imp_meas_future['Mangroves']['cost'][0]
            + cost_ben.imp_meas_future['Seawall']['cost'][0])
        self.assertAlmostEqual(new_cb.imp_meas_future[new_name]['cost'][1], 1)
        self.assertTrue(np.allclose(
            new_cb.imp_meas_future[new_name]['efc'].impact,
            new_cb.imp_meas_future[new_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_future[new_name]['risk_transf'], 0)

        self.assertAlmostEqual(new_cb.benefit[new_name], 51781337529.07264, places=3)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[new_name], 0.19679962474434248)

    def test_combine_current_pass(self):
        """Test combine_measures with only future"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018
        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, future_year=2040, risk_func=risk_aai_agg,
                      imp_time_depen=None, save_imp=True)

        new_name = 'combine'
        new_color = np.array([0.1, 0.1, 0.1])
        new_cb = cost_ben.combine_measures(['Mangroves', 'Seawall'], new_name, new_color,
                                           entity.disc_rates, imp_time_depen=None,
                                           risk_func=risk_aai_agg)

        self.assertTrue(np.allclose(new_cb.color_rgb[new_name], new_color))
        self.assertEqual(len(new_cb.imp_meas_present), 0)
        new_imp = cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event
        new_imp += cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Seawall']['impact'].at_event
        new_imp = np.maximum(cost_ben.imp_meas_future['no measure']['impact'].at_event - new_imp,
                             0)
        self.assertTrue(np.allclose(new_cb.imp_meas_future[new_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_future[new_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(
            new_cb.imp_meas_future[new_name]['cost'][0],
            cost_ben.imp_meas_future['Mangroves']['cost'][0]
            + cost_ben.imp_meas_future['Seawall']['cost'][0])
        self.assertAlmostEqual(new_cb.imp_meas_future[new_name]['cost'][1], 1)
        self.assertTrue(np.allclose(
            new_cb.imp_meas_future[new_name]['efc'].impact,
            new_cb.imp_meas_future[new_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_future[new_name]['risk_transf'], 0)
        self.assertAlmostEqual(new_cb.benefit[new_name], 51781337529.07264, places=3)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[new_name], 0.19679962474434248)

    def test_apply_transf_current_pass(self):
        """Test apply_risk_transfer with only future"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018
        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, future_year=2040, risk_func=risk_aai_agg,
                      imp_time_depen=None, save_imp=True)

        new_name = 'combine'
        new_color = np.array([0.1, 0.1, 0.1])
        risk_transf = (1.0e7, 15.0e11, 1)
        new_cb = cost_ben.combine_measures(['Mangroves', 'Seawall'], new_name, new_color,
                                           entity.disc_rates, imp_time_depen=None,
                                           risk_func=risk_aai_agg)
        new_cb.apply_risk_transfer(new_name, risk_transf[0], risk_transf[1],
                                   entity.disc_rates, cost_fix=0, cost_factor=risk_transf[2],
                                   imp_time_depen=1,
                                   risk_func=risk_aai_agg)

        tr_name = 'risk transfer (' + new_name + ')'
        new_imp = cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event
        new_imp += cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Seawall']['impact'].at_event
        new_imp = np.maximum(cost_ben.imp_meas_future['no measure']['impact'].at_event - new_imp,
                             0)
        imp_layer = np.minimum(np.maximum(new_imp - risk_transf[0], 0), risk_transf[1])
        risk_transfer = np.sum(
            imp_layer * cost_ben.imp_meas_future['no measure']['impact'].frequency)
        new_imp = np.maximum(new_imp - imp_layer, 0)

        self.assertTrue(np.allclose(new_cb.color_rgb[new_name], new_color))
        self.assertEqual(len(new_cb.imp_meas_present), 0)
        self.assertTrue(np.allclose(new_cb.imp_meas_future[tr_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_future[tr_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(
            new_cb.cost_ben_ratio[tr_name] * new_cb.benefit[tr_name],
            32106013195.316242, places=3)
        self.assertTrue(np.allclose(
            new_cb.imp_meas_future[tr_name]['efc'].impact,
            new_cb.imp_meas_future[tr_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_future[tr_name]['risk_transf'], risk_transfer)
        # benefit = impact layer
        self.assertAlmostEqual(new_cb.benefit[tr_name], 32106013195.316242, 4)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[tr_name], 1)

    def test_apply_transf_cost_fact_pass(self):
        """Test apply_risk_transfer with only future annd cost factor"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018
        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, future_year=2040, risk_func=risk_aai_agg,
                      imp_time_depen=None, save_imp=True)

        new_name = 'combine'
        new_color = np.array([0.1, 0.1, 0.1])
        risk_transf = (1.0e7, 15.0e11, 2)
        new_cb = cost_ben.combine_measures(['Mangroves', 'Seawall'], new_name, new_color,
                                           entity.disc_rates, imp_time_depen=None,
                                           risk_func=risk_aai_agg)
        new_cb.apply_risk_transfer(new_name, risk_transf[0], risk_transf[1],
                                   entity.disc_rates, cost_fix=0, cost_factor=risk_transf[2],
                                   imp_time_depen=1, risk_func=risk_aai_agg)

        tr_name = 'risk transfer (' + new_name + ')'
        new_imp = cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event
        new_imp += cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Seawall']['impact'].at_event
        new_imp = np.maximum(cost_ben.imp_meas_future['no measure']['impact'].at_event - new_imp,
                             0)
        imp_layer = np.minimum(np.maximum(new_imp - risk_transf[0], 0), risk_transf[1])
        risk_transfer = np.sum(
            imp_layer * cost_ben.imp_meas_future['no measure']['impact'].frequency)
        new_imp = np.maximum(new_imp - imp_layer, 0)

        self.assertTrue(np.allclose(new_cb.color_rgb[new_name], new_color))
        self.assertEqual(len(new_cb.imp_meas_present), 0)
        self.assertTrue(np.allclose(new_cb.imp_meas_future[tr_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_future[tr_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[tr_name] * new_cb.benefit[tr_name],
                               risk_transf[2] * 32106013195.316242)
        self.assertTrue(
            np.allclose(new_cb.imp_meas_future[tr_name]['efc'].impact,
                        new_cb.imp_meas_future[tr_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_future[tr_name]['risk_transf'], risk_transfer)
        # benefit = impact layer
        self.assertAlmostEqual(new_cb.benefit[tr_name], 32106013195.316242, 4)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[tr_name], risk_transf[2])

    def test_apply_transf_future_pass(self):
        """Test apply_risk_transfer with present and future"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018

        fut_ent = copy.deepcopy(entity)
        fut_ent.exposures.ref_year = 2040

        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, ent_future=fut_ent, risk_func=risk_aai_agg,
                      imp_time_depen=None, save_imp=True)

        new_name = 'combine'
        new_color = np.array([0.1, 0.1, 0.1])
        risk_transf = (1.0e7, 15.0e11, 1)
        new_cb = cost_ben.combine_measures(['Mangroves', 'Seawall'], new_name, new_color,
                                           entity.disc_rates, imp_time_depen=None,
                                           risk_func=risk_aai_agg)
        new_cb.apply_risk_transfer(new_name, risk_transf[0], risk_transf[1],
                                   entity.disc_rates, cost_fix=0, cost_factor=risk_transf[2],
                                   imp_time_depen=1, risk_func=risk_aai_agg)

        tr_name = 'risk transfer (' + new_name + ')'
        new_imp = cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Mangroves']['impact'].at_event
        new_imp += cost_ben.imp_meas_future['no measure']['impact'].at_event - \
            cost_ben.imp_meas_future['Seawall']['impact'].at_event
        new_imp = np.maximum(cost_ben.imp_meas_future['no measure']['impact'].at_event - new_imp,
                             0)
        imp_layer = np.minimum(np.maximum(new_imp - risk_transf[0], 0), risk_transf[1])
        risk_transfer = np.sum(
            imp_layer * cost_ben.imp_meas_future['no measure']['impact'].frequency)
        new_imp = np.maximum(new_imp - imp_layer, 0)

        self.assertTrue(np.allclose(new_cb.color_rgb[new_name], new_color))
        self.assertEqual(len(new_cb.imp_meas_present), 3)
        self.assertTrue(np.allclose(new_cb.imp_meas_future[tr_name]['impact'].at_event, new_imp))
        self.assertTrue(np.allclose(new_cb.imp_meas_present[tr_name]['impact'].at_event, new_imp))
        self.assertAlmostEqual(
            new_cb.imp_meas_future[tr_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(
            new_cb.imp_meas_present[tr_name]['risk'],
            np.sum(new_imp * cost_ben.imp_meas_future['no measure']['impact'].frequency), 5)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[tr_name] * new_cb.benefit[tr_name],
                               69715165679.7042, places=3)
        self.assertTrue(
            np.allclose(new_cb.imp_meas_future[tr_name]['efc'].impact,
                        new_cb.imp_meas_future[tr_name]['impact'].calc_freq_curve().impact))
        self.assertAlmostEqual(new_cb.imp_meas_future[tr_name]['risk_transf'], risk_transfer)
        # benefit = impact layer
        self.assertAlmostEqual(new_cb.benefit[tr_name], 69715165679.7042, 4)
        self.assertAlmostEqual(new_cb.cost_ben_ratio[tr_name], 1)

    def test_remove_measure(self):
        """Test remove_measure method"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018
        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, future_year=2040, risk_func=risk_aai_agg,
                      imp_time_depen=None, save_imp=True)

        to_remove = 'Mangroves'
        self.assertTrue(to_remove in cost_ben.benefit.keys())
        cost_ben.remove_measure(to_remove)
        self.assertTrue(to_remove not in cost_ben.color_rgb.keys())
        self.assertTrue(to_remove not in cost_ben.benefit.keys())
        self.assertTrue(to_remove not in cost_ben.cost_ben_ratio.keys())
        self.assertTrue(to_remove not in cost_ben.imp_meas_future.keys())
        self.assertTrue(to_remove not in cost_ben.imp_meas_present.keys())
        self.assertEqual(len(cost_ben.imp_meas_present), 0)
        self.assertEqual(len(cost_ben.imp_meas_future), 4)
        self.assertEqual(len(cost_ben.color_rgb), 4)
        self.assertEqual(len(cost_ben.cost_ben_ratio), 3)
        self.assertEqual(len(cost_ben.benefit), 3)

class TestCalc(unittest.TestCase):
    """Test calc"""

    def test_calc_change_pass(self):
        """Test calc with future change"""
        # present
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.exposures.gdf.rename(columns={'impf_': 'impf_TC'}, inplace=True)
        entity.check()
        entity.exposures.ref_year = 2018

        # future
        ent_future = Entity.from_excel(ENT_DEMO_FUTURE)
        ent_future.check()
        ent_future.exposures.ref_year = 2040

        haz_future = copy.deepcopy(hazard)
        haz_future.intensity.data += 25

        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, haz_future, ent_future)

        self.assertEqual(cost_ben.present_year, 2018)
        self.assertEqual(cost_ben.future_year, 2040)
        self.assertAlmostEqual(cost_ben.tot_climate_risk, 5.768659152882021e+11, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_present['no measure']['risk'],
                               6.51220115756442e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Mangroves']['risk'],
                               4.850407096284983e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Beach nourishment']['risk'],
                               5.188921355413834e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Seawall']['risk'],
                               4.736400526119911e+09, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_present['Building code']['risk'],
                               4.884150868173321e+09, places=3)

        self.assertAlmostEqual(cost_ben.imp_meas_future['no measure']['risk'],
                               5.9506659786664024e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Mangroves']['risk'],
                               4.826231151473135e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Beach nourishment']['risk'],
                               5.0647250923231674e+10, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Seawall']['risk'],
                               21089567135.7345, places=3)
        self.assertAlmostEqual(cost_ben.imp_meas_future['Building code']['risk'],
                               4.462999483999791e+10, places=3)

        self.assertAlmostEqual(cost_ben.benefit['Mangroves'], 113345027690.81276, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Beach nourishment'], 89444869971.53653, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Seawall'], 347977469896.1333, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Building code'], 144216478822.05154, places=3)

        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Mangroves'], 0.011573232523528404)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Beach nourishment'], 0.01931916274851638)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Seawall'], 0.025515385913577368)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Building code'], 0.06379298728650741)

        self.assertAlmostEqual(cost_ben.tot_climate_risk, 576865915288.2021, places=3)

    def test_calc_no_change_pass(self):
        """Test calc without future change"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        entity = Entity.from_excel(ENT_DEMO_TODAY)
        entity.check()
        entity.exposures.ref_year = 2018
        cost_ben = CostBenefit()
        cost_ben.calc(hazard, entity, future_year=2040)

        self.assertEqual(cost_ben.imp_meas_present, dict())
        self.assertEqual(len(cost_ben.imp_meas_future), 5)
        self.assertEqual(cost_ben.present_year, 2018)
        self.assertEqual(cost_ben.future_year, 2040)

        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Mangroves'], 0.04230714690616641)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Beach nourishment'], 0.06998836431681373)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Seawall'], 0.2679741183248266)
        self.assertAlmostEqual(cost_ben.cost_ben_ratio['Building code'], 0.30286828677985717)

        self.assertAlmostEqual(cost_ben.benefit['Mangroves'], 3.100583368954022e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Beach nourishment'],
                               2.468981832719974e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Seawall'], 3.3132973770502796e+10, places=3)
        self.assertAlmostEqual(cost_ben.benefit['Building code'], 3.0376240767284798e+10, places=3)

        self.assertAlmostEqual(cost_ben.tot_climate_risk, 1.2150496306913972e+11, places=3)

class TestRiskFuncs(unittest.TestCase):
    """Test risk functions definitions"""

    def test_impact(self):
        ent = Entity.from_excel(ENT_DEMO_TODAY)
        ent.check()
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        impact = Impact()
        ent.exposures.assign_centroids(hazard)
        impact.calc(ent.exposures, ent.impact_funcs, hazard)
        return impact

    def test_risk_aai_agg_pass(self):
        """Test risk_aai_agg"""
        impact = self.test_impact()
        risk = risk_aai_agg(impact)
        self.assertAlmostEqual(6.512201157564421e+09, risk, 5)
        self.assertTrue(np.isclose(6.512201157564421e+09, risk))

    def test_risk_rp_100_pass(self):
        """Test risk_rp_100"""
        impact = self.test_impact()
        exc_freq = impact.calc_freq_curve([100])

        risk = risk_rp_100(impact)
        self.assertAlmostEqual(exc_freq.impact[0], risk)

    def test_risk_rp_200_pass(self):
        """Test risk_rp_200"""
        impact = self.test_impact()
        exc_freq = impact.calc_freq_curve([250])

        risk = risk_rp_250(impact)
        self.assertAlmostEqual(exc_freq.impact[0], risk)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestRiskFuncs)
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestCalc)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSteps))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRiskFuncs))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

module containing functions to check variables properties.
"""

__all__ = [
    'size',
    'shape',
    'array_optional',
    'array_default'
]

import logging
import numpy as np
import scipy.sparse as sparse

LOGGER = logging.getLogger(__name__)

def check_oligatories(var_dict, var_obl, name_prefix, n_size, n_row, n_col):
    """Check size of obligatory variables.

    Paraemters:
        var_dict (dict): __dict__ class attribute
        var_obl (set): name of the obligatory variables
        name_prefix (str): name to add in the error log, e.g. the class name
        n_size (int): size expected from arrays and lists
        n_row (int): number of rows expected in 2D arrays
        n_col (int): number of columns expected in 2D arrays

    Raises
    ------
    ValueError
    """
    for var_name, var_val in var_dict.items():
        if var_name in var_obl:
            if (isinstance(var_val, np.ndarray) and var_val.ndim == 1) \
               or isinstance(var_val, list):
                size(n_size, var_val, name_prefix + var_name)
            elif (isinstance(var_val, np.ndarray) and var_val.ndim == 2):
                shape(n_row, n_col, var_val, name_prefix + var_name)
            elif isinstance(var_val, (np.ndarray, sparse.csr.csr_matrix)) and var_val.ndim == 2:
                shape(n_row, n_col, var_val, name_prefix + var_name)

def check_optionals(var_dict, var_opt, name_prefix, n_size):
    """Check size of obligatory variables.

    Paraemters:
        var_dict (dict): __dict__ class attribute
        var_opt (set): name of the ooptional variables
        name_prefix (str): name to add in the error log, e.g. the class name
        n_size (int): size expected from arrays and lists

    Raises
    ------
    ValueError
    """
    for var_name, var_val in var_dict.items():
        if var_name in var_opt:
            if isinstance(var_val, (np.ndarray, list)):
                array_optional(n_size, var_val, name_prefix + var_name)

def empty_optional(var, var_name):
    """Check if a data structure is empty."""
    if not var:
        LOGGER.debug("%s not set. ", var_name)

def size(exp_len, var, var_name):
    """Check if the length of a variable is the expected one.

        Raises
        ------
        ValueError
    """
    try:
        if isinstance(exp_len, int):
            if exp_len != len(var):
                raise ValueError(f"Invalid {var_name} size: {str(exp_len)} != {len(var)}.")
        elif len(var) not in exp_len:
            raise ValueError(f"Invalid {var_name} size: {len(var)} not in {str(exp_len)}.")
    except TypeError as err:
        raise ValueError(f"{var_name} has wrong size.") from err

def shape(exp_row, exp_col, var, var_name):
    """Check if the length of a variable is the expected one.

        Raises
        ------
        ValueError
    """
    try:
        if exp_row != var.shape[0]:
            raise ValueError(f"Invalid {var_name} row size: {exp_row} != {var.shape[0]}.")
        if exp_col != var.shape[1]:
            raise ValueError(f"Invalid {var_name} column size: {exp_col} != {var.shape[1]}.")
    except TypeError as err:
        raise ValueError("%s has wrong dimensions." % var_name) from err


def array_optional(exp_len, var, var_name):
    """Check if array has right size. Warn if array empty. Call check_size.

        Parameters
        ----------
        exp_len : str
            expected array size
        var : np.array
            numpy array to check
        var_name : str
            name of the variable. Used in error/warning msg

        Raises
        ------
        ValueError
    """
    if len(var) == 0 and exp_len > 0:
        LOGGER.debug("%s not set. ", var_name)
    else:
        size(exp_len, var, var_name)

def array_default(exp_len, var, var_name, def_val):
    """Check array has right size. Set default value if empty. Call check_size.

        Parameters
        ----------
        exp_len : str
            expected array size
        var : np.array
            numpy array to check
        var_name : str
            name of the variable. Used in error/warning msg
        def_val : np.array
            nump array used as default value

        Raises
        ------
        ValueError

        Returns
        -------
        Filled array
    """
    res = var
    if len(var) == 0 and exp_len > 0:
        LOGGER.debug("%s not set. Default values set.", var_name)
        res = def_val
    else:
        size(exp_len, var, var_name)
    return res



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define TC wind hazard (TropCyclone class).
"""

__all__ = ['TropCyclone']

import itertools
import logging
import copy
import time
import datetime as dt
import numpy as np
from scipy import sparse
import matplotlib.animation as animation
from tqdm import tqdm

from climada.hazard.base import Hazard
from climada.hazard.tag import Tag as TagHazard
from climada.hazard.tc_tracks import TCTracks, estimate_rmw
from climada.hazard.tc_clim_change import get_knutson_criterion, calc_scale_knutson
from climada.hazard.centroids.centr import Centroids
from climada.util import ureg
import climada.util.coordinates as u_coord
import climada.util.plot as u_plot

LOGGER = logging.getLogger(__name__)

HAZ_TYPE = 'TC'
"""Hazard type acronym for Tropical Cyclone"""

INLAND_MAX_DIST_KM = 1000
"""Maximum inland distance of the centroids in km"""

CENTR_NODE_MAX_DIST_KM = 300
"""Maximum distance between centroid and TC track node in km"""

CENTR_NODE_MAX_DIST_DEG = 5.5
"""Maximum distance between centroid and TC track node in degrees"""

DEF_INTENSITY_THRES = 17.5
"""Wind speeds (in m/s) below this threshold are stored as 0 if no other threshold is specified."""

MODEL_VANG = {'H08': 0, 'H1980': 1, 'H10': 2}
"""Enumerate different symmetric wind field models."""

RHO_AIR = 1.15
"""Air density. Assumed constant, following Holland 1980."""

GRADIENT_LEVEL_TO_SURFACE_WINDS = 0.9
"""Gradient-to-surface wind reduction factor according to the 90%-rule:

Franklin, J.L., Black, M.L., Valde, K. (2003): GPS Dropwindsonde Wind Profiles in Hurricanes and
Their Operational Implications. Weather and Forecasting 18(1): 32–44.
https://doi.org/10.1175/1520-0434(2003)018<0032:GDWPIH>2.0.CO;2
"""

KMH_TO_MS = (1.0 * ureg.km / ureg.hour).to(ureg.meter / ureg.second).magnitude
KN_TO_MS = (1.0 * ureg.knot).to(ureg.meter / ureg.second).magnitude
NM_TO_KM = (1.0 * ureg.nautical_mile).to(ureg.kilometer).magnitude
"""Unit conversion factors for JIT functions that can't use ureg"""

class TropCyclone(Hazard):
    """
    Contains tropical cyclone events.

    Attributes
    ----------
    category : np.array(int)
        for every event, the TC category using the
        Saffir-Simpson scale:
            -1 tropical depression
             0 tropical storm
             1 Hurrican category 1
             2 Hurrican category 2
             3 Hurrican category 3
             4 Hurrican category 4
             5 Hurrican category 5
    basin : list(str)
        basin where every event starts
        'NA' North Atlantic
        'EP' Eastern North Pacific
        'WP' Western North Pacific
        'NI' North Indian
        'SI' South Indian
        'SP' Southern Pacific
        'SA' South Atlantic
    """
    intensity_thres = DEF_INTENSITY_THRES
    """intensity threshold for storage in m/s"""

    vars_opt = Hazard.vars_opt.union({'category'})
    """Name of the variables that aren't need to compute the impact."""

    def __init__(self, pool=None):
        """Initialize values.

        Parameters
        ----------
        pool : pathos.pool, optional
            Pool that will be used for parallel computation when applicable. Default: None
        """
        Hazard.__init__(self, haz_type=HAZ_TYPE, pool=pool)
        self.category = np.array([], int)
        self.basin = []
        self.windfields = []

    def set_from_tracks(self, *args, **kwargs):
        """This function is deprecated, use TropCyclone.from_tracks instead."""
        LOGGER.warning("The use of TropCyclone.set_from_tracks is deprecated."
                       "Use TropCyclone.from_tracks instead.")
        if "intensity_thres" not in kwargs:
            # some users modify the threshold attribute before calling `set_from_tracks`
            kwargs["intensity_thres"] = self.intensity_thres
        if self.pool is not None and 'pool' not in kwargs:
            kwargs['pool'] = self.pool
        self.__dict__ = TropCyclone.from_tracks(*args, **kwargs).__dict__

    @classmethod
    def from_tracks(cls, tracks, centroids=None, pool=None, description='', model='H08',
                    ignore_distance_to_coast=False, store_windfields=False, metric="equirect",
                    intensity_thres=DEF_INTENSITY_THRES):
        """
        Create new TropCyclone instance that contains windfields from the specified tracks.

        This function sets the `intensity` attribute to contain, for each centroid,
        the maximum wind speed (1-minute sustained winds at 10 meters above ground) experienced
        over the whole period of each TC event in m/s. The wind speed is set to 0 if it doesn't
        exceed the threshold `intensity_thres`.

        The `category` attribute is set to the value of the `category`-attribute
        of each of the given track data sets.

        The `basin` attribute is set to the genesis basin for each event, which
        is the first value of the `basin`-variable in each of the given track data sets.

        Optionally, the time dependent, vectorial winds can be stored using the `store_windfields`
        function parameter (see below).

        Parameters
        ----------
        tracks : TCTracks
            Tracks of storm events.
        centroids : Centroids, optional
            Centroids where to model TC. Default: global centroids at 360 arc-seconds resolution.
        pool : pathos.pool, optional
            Pool that will be used for parallel computation of wind fields. Default: None
        description : str, optional
            Description of the event set. Default: "".
        model : str, optional
            Parametric wind field model to use: one of "H1980" (the prominent Holland 1980 model),
            "H08" (Holland 1980 with b-value from Holland 2008), or "H10" (Holland et al. 2010).
            Default: "H08".
        ignore_distance_to_coast : boolean, optional
            If True, centroids far from coast are not ignored. Default: False.
        store_windfields : boolean, optional
            If True, the Hazard object gets a list `windfields` of sparse matrices. For each track,
            the full velocity vectors at each centroid and track position are stored in a sparse
            matrix of shape (npositions,  ncentroids * 2) that can be reshaped to a full ndarray
            of shape (npositions, ncentroids, 2). Default: False.
        metric : str, optional
            Specify an approximation method to use for earth distances:
            * "equirect": Distance according to sinusoidal projection. Fast, but inaccurate for
              large distances and high latitudes.
            * "geosphere": Exact spherical distance. Much more accurate at all distances, but slow.
            Default: "equirect".
        intensity_thres : float, optional
            Wind speeds (in m/s) below this threshold are stored as 0. Default: 17.5

        Raises
        ------
        ValueError

        Returns
        -------
        TropCyclone
        """
        num_tracks = tracks.size
        if centroids is None:
            centroids = Centroids.from_base_grid(res_as=360, land=False)

        if not centroids.coord.size:
            centroids.set_meta_to_lat_lon()

        if ignore_distance_to_coast:
            # Select centroids with lat < 61
            coastal_idx = (np.abs(centroids.lat) < 61).nonzero()[0]
        else:
            # Select centroids which are inside INLAND_MAX_DIST_KM and lat < 61
            if not centroids.dist_coast.size:
                centroids.set_dist_coast()
            coastal_idx = ((centroids.dist_coast < INLAND_MAX_DIST_KM * 1000)
                           & (np.abs(centroids.lat) < 61)).nonzero()[0]

        # Restrict to coastal centroids within reach of any of the tracks
        t_lon_min, t_lat_min, t_lon_max, t_lat_max = tracks.get_bounds(
            deg_buffer=CENTR_NODE_MAX_DIST_DEG)
        t_mid_lon = 0.5 * (t_lon_min + t_lon_max)
        coastal_centroids = centroids.coord[coastal_idx]
        u_coord.lon_normalize(coastal_centroids[:, 1], center=t_mid_lon)
        coastal_idx = coastal_idx[((t_lon_min <= coastal_centroids[:, 1])
                                   & (coastal_centroids[:, 1] <= t_lon_max)
                                   & (t_lat_min <= coastal_centroids[:, 0])
                                   & (coastal_centroids[:, 0] <= t_lat_max))]

        LOGGER.info('Mapping %s tracks to %s coastal centroids.', str(tracks.size),
                    str(coastal_idx.size))
        if pool:
            chunksize = min(num_tracks // pool.ncpus, 1000)
            tc_haz_list = pool.map(
                cls.from_single_track, tracks.data,
                itertools.repeat(centroids, num_tracks),
                itertools.repeat(coastal_idx, num_tracks),
                itertools.repeat(model, num_tracks),
                itertools.repeat(store_windfields, num_tracks),
                itertools.repeat(metric, num_tracks),
                itertools.repeat(intensity_thres, num_tracks),
                chunksize=chunksize)
        else:
            last_perc = 0
            tc_haz_list = []
            for track in tracks.data:
                perc = 100 * len(tc_haz_list) / len(tracks.data)
                if perc - last_perc >= 10:
                    LOGGER.info("Progress: %d%%", perc)
                    last_perc = perc
                tc_haz_list.append(
                    cls.from_single_track(track, centroids, coastal_idx,
                                          model=model, store_windfields=store_windfields,
                                          metric=metric, intensity_thres=intensity_thres))
            if last_perc < 100:
                LOGGER.info("Progress: 100%")

        LOGGER.debug('Concatenate events.')
        haz = cls.concat(tc_haz_list)
        haz.pool = pool
        haz.intensity_thres = intensity_thres
        LOGGER.debug('Compute frequency.')
        haz.frequency_from_tracks(tracks.data)
        haz.tag.description = description
        return haz

    def apply_climate_scenario_knu(self, ref_year=2050, rcp_scenario=45):
        """
        From current TC hazard instance, return new hazard set with
        future events for a given RCP scenario and year based on the
        parametrized values derived from Table 3 in Knutson et al 2015.
        https://doi.org/10.1175/JCLI-D-15-0129.1 . The scaling for different
        years and RCP scenarios is obtained by linear interpolation.

        Note: The parametrized values are derived from the overall changes
        in statistical ensemble of tracks. Hence, this method should only be
        applied to sufficiently large tropical cyclone event sets that
        approximate the reference years 1981 - 2008 used in Knutson et. al.

        The frequency and intensity changes are applied independently from
        one another. The mean intensity factors can thus slightly deviate
        from the Knutson value (deviation was found to be less than 1%
        for default IBTrACS event sets 1980-2020 for each basin).

        Parameters
        ----------
        ref_year : int
            year between 2000 ad 2100. Default: 2050
        rcp_scenario : int
            26 for RCP 2.6, 45 for RCP 4.5, 60 for RCP 6.0 and 85 for RCP 8.5.
            The default is 45.

        Returns
        -------
        haz_cc : climada.hazard.TropCyclone
            Tropical cyclone with frequencies and intensity scaled according
            to the Knutson criterion for the given year and RCP. Returns
            a new instance of climada.hazard.TropCyclone, self is not
            modified.
        """
        chg_int_freq = get_knutson_criterion()
        scale_rcp_year  = calc_scale_knutson(ref_year, rcp_scenario)
        haz_cc = self._apply_knutson_criterion(chg_int_freq, scale_rcp_year)
        haz_cc.tag.description = 'climate change scenario for year %s and RCP %s '\
        'from Knutson et al 2015.' % (str(ref_year), str(rcp_scenario))
        return haz_cc

    def set_climate_scenario_knu(self, *args, **kwargs):
        """This function is deprecated, use TropCyclone.apply_climate_scenario_knu instead."""
        LOGGER.warning("The use of TropCyclone.set_climate_scenario_knu is deprecated."
                       "Use TropCyclone.apply_climate_scenario_knu instead.")
        return self.apply_climate_scenario_knu(*args, **kwargs)

    @classmethod
    def video_intensity(cls, track_name, tracks, centroids, file_name=None,
                        writer=animation.PillowWriter(bitrate=500),
                        figsize=(9, 13), adapt_fontsize=True, **kwargs):
        """
        Generate video of TC wind fields node by node and returns its
        corresponding TropCyclone instances and track pieces.

        Parameters
        ----------
        track_name : str
            name of the track contained in tracks to record
        tracks : climada.hazard.TCTracks
            tropical cyclone tracks
        centroids : climada.hazard.Centroids
            centroids where wind fields are mapped
        file_name : str, optional
            file name to save video (including full path and file extension)
        writer : matplotlib.animation.*, optional
            video writer. Default is pillow with bitrate=500
        figsize : tuple, optional
            figure size for plt.subplots
        adapt_fontsize : bool, optional
            If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
            the default matplotlib font size is used. Default is True.
        kwargs : optional
            arguments for pcolormesh matplotlib function used in event plots

        Returns
        -------
        tc_list, tc_coord : list(TropCyclone), list(np.array)

        Raises
        ------
        ValueError

        """
        # initialization
        track = tracks.get_track(track_name)
        if not track:
            raise ValueError('%s not found in track data.' % track_name)
        idx_plt = np.argwhere(
            (track.lon.values < centroids.total_bounds[2] + 1)
            & (centroids.total_bounds[0] - 1 < track.lon.values)
            & (track.lat.values < centroids.total_bounds[3] + 1)
            & (centroids.total_bounds[1] - 1 < track.lat.values)
        ).reshape(-1)

        tc_list = []
        tr_coord = {'lat': [], 'lon': []}
        for node in range(idx_plt.size - 2):
            tr_piece = track.sel(
                time=slice(track.time.values[idx_plt[node]],
                           track.time.values[idx_plt[node + 2]]))
            tr_piece.attrs['n_nodes'] = 2  # plot only one node
            tr_sel = TCTracks()
            tr_sel.append(tr_piece)
            tr_coord['lat'].append(tr_sel.data[0].lat.values[:-1])
            tr_coord['lon'].append(tr_sel.data[0].lon.values[:-1])

            tc_tmp = cls.from_tracks(tr_sel, centroids=centroids)
            tc_tmp.event_name = [
                track.name + ' ' + time.strftime(
                    "%d %h %Y %H:%M",
                    time.gmtime(tr_sel.data[0].time[1].values.astype(int)
                                / 1000000000)
                )
            ]
            tc_list.append(tc_tmp)

        if 'cmap' not in kwargs:
            kwargs['cmap'] = 'Greys'
        if 'vmin' not in kwargs:
            kwargs['vmin'] = np.array([tc_.intensity.min() for tc_ in tc_list]).min()
        if 'vmax' not in kwargs:
            kwargs['vmax'] = np.array([tc_.intensity.max() for tc_ in tc_list]).max()

        def run(node):
            tc_list[node].plot_intensity(1, axis=axis, **kwargs)
            axis.plot(tr_coord['lon'][node], tr_coord['lat'][node], 'k')
            axis.set_title(tc_list[node].event_name[0])
            pbar.update()

        if file_name:
            LOGGER.info('Generating video %s', file_name)
            fig, axis, fontsize = u_plot.make_map(figsize=figsize, adapt_fontsize=adapt_fontsize)
            pbar = tqdm(total=idx_plt.size - 2)
            ani = animation.FuncAnimation(fig, run, frames=idx_plt.size - 2,
                                          interval=500, blit=False)
            fig.tight_layout()
            ani.save(file_name, writer=writer)
            pbar.close()
        return tc_list, tr_coord

    def frequency_from_tracks(self, tracks):
        """
        Set hazard frequency from tracks data.

        Parameters
        ----------
        tracks : list of xarray.Dataset
        """
        if not tracks:
            return
        year_max = np.amax([t.time.dt.year.values.max() for t in tracks])
        year_min = np.amin([t.time.dt.year.values.min() for t in tracks])
        year_delta = year_max - year_min + 1
        num_orig = np.count_nonzero(self.orig)
        ens_size = (self.event_id.size / num_orig) if num_orig > 0 else 1
        self.frequency = np.ones(self.event_id.size) / (year_delta * ens_size)

    @classmethod
    def from_single_track(cls, track, centroids, coastal_idx, model='H08',
                          store_windfields=False, metric="equirect",
                          intensity_thres=DEF_INTENSITY_THRES):
        """
        Generate windfield hazard from a single track dataset

        Parameters
        ----------
        track : xr.Dataset
            Single tropical cyclone track.
        centroids : Centroids
            Centroids instance.
        coastal_idx : np.array
            Indices of centroids close to coast.
        model : str, optional
            Parametric wind field model, one of "H1980" (the prominent Holland 1980 model),
            "H08" (Holland 1980 with b-value from Holland 2008), or "H10" (Holland et al. 2010).
            Default: "H08".
        store_windfields : boolean, optional
            If True, store windfields. Default: False.
        metric : str, optional
            Specify an approximation method to use for earth distances: "equirect" (faster) or
            "geosphere" (more accurate). See `dist_approx` function in `climada.util.coordinates`.
            Default: "equirect".
        intensity_thres : float, optional
            Wind speeds (in m/s) below this threshold are stored as 0. Default: 17.5

        Raises
        ------
        ValueError, KeyError

        Returns
        -------
        haz : TropCyclone
        """
        try:
            mod_id = MODEL_VANG[model]
        except KeyError as err:
            raise ValueError(f'Model not implemented: {model}.') from err
        ncentroids = centroids.coord.shape[0]
        coastal_centr = centroids.coord[coastal_idx]
        windfields, reachable_centr_idx = compute_windfields(track, coastal_centr, mod_id,
                                                             metric=metric)
        reachable_coastal_centr_idx = coastal_idx[reachable_centr_idx]
        npositions = windfields.shape[0]

        intensity = np.linalg.norm(windfields, axis=-1).max(axis=0)
        intensity[intensity < intensity_thres] = 0
        intensity_sparse = sparse.csr_matrix(
            (intensity, reachable_coastal_centr_idx, [0, intensity.size]),
            shape=(1, ncentroids))
        intensity_sparse.eliminate_zeros()

        new_haz = cls()
        new_haz.tag = TagHazard(HAZ_TYPE, 'Name: ' + track.name)
        new_haz.intensity_thres = intensity_thres
        new_haz.intensity = intensity_sparse
        if store_windfields:
            n_reachable_coastal_centr = reachable_coastal_centr_idx.size
            indices = np.zeros((npositions, n_reachable_coastal_centr, 2), dtype=np.int64)
            indices[:, :, 0] = 2 * reachable_coastal_centr_idx[None]
            indices[:, :, 1] = 2 * reachable_coastal_centr_idx[None] + 1
            indices = indices.ravel()
            indptr = np.arange(npositions + 1) * n_reachable_coastal_centr * 2
            windfields_sparse = sparse.csr_matrix((windfields.ravel(), indices, indptr),
                                                  shape=(npositions, ncentroids * 2))
            windfields_sparse.eliminate_zeros()
            new_haz.windfields = [windfields_sparse]
        new_haz.units = 'm/s'
        new_haz.centroids = centroids
        new_haz.event_id = np.array([1])
        new_haz.frequency = np.array([1])
        new_haz.event_name = [track.sid]
        new_haz.fraction = new_haz.intensity.copy()
        new_haz.fraction.data.fill(1)
        # store first day of track as date
        new_haz.date = np.array([
            dt.datetime(track.time.dt.year.values[0],
                        track.time.dt.month.values[0],
                        track.time.dt.day.values[0]).toordinal()
        ])
        new_haz.orig = np.array([track.orig_event_flag])
        new_haz.category = np.array([track.category])
        # users that pickle TCTracks objects might still have data with the legacy basin attribute,
        # so we have to deal with it here
        new_haz.basin = [track.basin if isinstance(track.basin, str)
                         else str(track.basin.values[0])]
        return new_haz

    def _apply_knutson_criterion(self, chg_int_freq, scaling_rcp_year):
        """
        Apply changes to intensities and cumulative frequencies.

        Parameters
        ----------
        criterion : list(dict))
            list of criteria from climada.hazard.tc_clim_change
        scale : float
            scale parameter because of chosen year and RCP
        Returns
        -------
        tc_cc : climada.hazard.TropCyclone
            Tropical cyclone with frequency and intensity scaled inspired by
            the Knutson criterion. Returns a new instance of TropCyclone.
        """

        tc_cc = copy.deepcopy(self)

        # Criterion per basin
        for basin in np.unique(tc_cc.basin):

            bas_sel = (np.array(tc_cc.basin) == basin)

            # Apply intensity change
            inten_chg = [chg
                         for chg in chg_int_freq
                         if (chg['variable'] == 'intensity' and
                             chg['basin'] == basin)
                         ]
            for chg in inten_chg:
                sel_cat_chg = np.isin(tc_cc.category, chg['category']) & bas_sel
                inten_scaling = 1 + (chg['change'] - 1) * scaling_rcp_year
                tc_cc.intensity = sparse.diags(
                    np.where(sel_cat_chg, inten_scaling, 1)
                    ).dot(tc_cc.intensity)

            # Apply frequency change
            freq_chg = [chg
                        for chg in chg_int_freq
                        if (chg['variable'] == 'frequency' and
                            chg['basin'] == basin)
                        ]
            freq_chg.sort(reverse=False, key=lambda x: len(x['category']))

            # Scale frequencies by category
            cat_larger_list = []
            for chg in freq_chg:
                cat_chg_list = [cat
                                for cat in chg['category']
                                if cat not in cat_larger_list
                                ]
                sel_cat_chg = np.isin(tc_cc.category, cat_chg_list) & bas_sel
                if sel_cat_chg.any():
                    freq_scaling = 1 + (chg['change'] - 1) * scaling_rcp_year
                    tc_cc.frequency[sel_cat_chg] *= freq_scaling
                cat_larger_list += cat_chg_list

        if (tc_cc.frequency < 0).any():
            raise ValueError("The application of the given climate scenario"
                             "resulted in at least one negative frequency.")

        return tc_cc


def compute_windfields(track, centroids, model, metric="equirect"):
    """Compute 1-minute sustained winds (in m/s) at 10 meters above ground

    In a first step, centroids within reach of the track are determined so that wind fields will
    only be computed and returned for those centroids.

    Parameters
    ----------
    track : xr.Dataset
        Track infomation.
    centroids : 2d np.array
        Each row is a centroid [lat, lon].
        Centroids that are not within reach of the track are ignored.
    model : int
        Holland model selection according to MODEL_VANG.
    metric : str, optional
        Specify an approximation method to use for earth distances: "equirect" (faster) or
        "geosphere" (more accurate). See `dist_approx` function in `climada.util.coordinates`.
        Default: "equirect".

    Returns
    -------
    windfields : np.array of shape (npositions, nreachable, 2)
        Directional wind fields for each track position on those centroids within reach
        of the TC track.
    reachable_centr_idx : np.array of shape (nreachable,)
        List of indices of input centroids within reach of the TC track.
    """
    # copies of track data
    # Note that max wind records are not used in the Holland wind field models!
    t_lat, t_lon, t_tstep, t_rad, t_env, t_cen = [
        track[ar].values.copy() for ar in ['lat', 'lon', 'time_step', 'radius_max_wind',
                                           'environmental_pressure', 'central_pressure']
    ]

    # start with the assumption that no centroids are within reach
    npositions = t_lat.shape[0]
    reachable_centr_idx = np.zeros((0,), dtype=np.int64)
    windfields = np.zeros((npositions, 0, 2), dtype=np.float64)

    # the wind field model requires at least two track positions because translational speed
    # as well as the change in pressure are required
    if npositions < 2:
        return windfields, reachable_centr_idx

    # normalize longitude values (improves performance of `dist_approx` and `_close_centroids`)
    mid_lon = 0.5 * sum(u_coord.lon_bounds(t_lon))
    u_coord.lon_normalize(t_lon, center=mid_lon)
    u_coord.lon_normalize(centroids[:, 1], center=mid_lon)

    # restrict to centroids within rectangular bounding boxes around track positions
    track_centr_msk = _close_centroids(t_lat, t_lon, centroids)
    track_centr = centroids[track_centr_msk]
    nreachable = track_centr.shape[0]
    if nreachable == 0:
        return windfields, reachable_centr_idx

    # compute distances and vectors to all centroids
    [d_centr], [v_centr] = u_coord.dist_approx(t_lat[None], t_lon[None],
                                               track_centr[None, :, 0], track_centr[None, :, 1],
                                               log=True, normalize=False, method=metric)

    # exclude centroids that are too far from or too close to the eye
    close_centr_msk = (d_centr < CENTR_NODE_MAX_DIST_KM) & (d_centr > 1e-2)
    if not np.any(close_centr_msk):
        return windfields, reachable_centr_idx
    v_centr_normed = np.zeros_like(v_centr)
    v_centr_normed[close_centr_msk] = v_centr[close_centr_msk] / d_centr[close_centr_msk, None]

    # make sure that central pressure never exceeds environmental pressure
    pres_exceed_msk = (t_cen > t_env)
    t_cen[pres_exceed_msk] = t_env[pres_exceed_msk]

    # extrapolate radius of max wind from pressure if not given
    t_rad[:] = estimate_rmw(t_rad, t_cen) * NM_TO_KM

    # translational speed of track at every node
    v_trans = _vtrans(t_lat, t_lon, t_tstep, metric=metric)
    v_trans_norm = v_trans[0]

    # adjust pressure at previous track point
    prev_pres = t_cen[:-1].copy()
    msk = (prev_pres < 850)
    prev_pres[msk] = t_cen[1:][msk]

    # compute b-value and derive (absolute) angular velocity
    if model == MODEL_VANG['H1980']:
        # convert recorded surface winds to gradient-level winds without translational influence
        t_vmax = track.max_sustained_wind.values.copy()
        t_gradient_winds = np.fmax(0, t_vmax - v_trans_norm) / GRADIENT_LEVEL_TO_SURFACE_WINDS
        hol_b = _B_holland_1980(t_gradient_winds[1:], t_env[1:], t_cen[1:])
        v_ang_norm = _stat_holland_1980(d_centr[1:], t_rad[1:], hol_b, t_env[1:],
                                        t_cen[1:], t_lat[1:], close_centr_msk[1:])
        v_ang_norm *= GRADIENT_LEVEL_TO_SURFACE_WINDS
    elif model == MODEL_VANG['H08']:
        # this model doesn't use the recorded surface winds
        hol_b = _bs_holland_2008(v_trans_norm[1:], t_env[1:], t_cen[1:], prev_pres,
                                 t_lat[1:], t_tstep[1:])
        v_ang_norm = _stat_holland_1980(d_centr[1:], t_rad[1:], hol_b, t_env[1:],
                                        t_cen[1:], t_lat[1:], close_centr_msk[1:])
    elif model == MODEL_VANG['H10']:
        # this model doesn't use the recorded surface winds
        hol_b = _bs_holland_2008(v_trans_norm[1:], t_env[1:], t_cen[1:], prev_pres,
                                 t_lat[1:], t_tstep[1:])
        t_vmax = _v_max_s_holland_2008(t_env[1:], t_cen[1:], hol_b)
        hol_x = _x_holland_2010(d_centr[1:], t_rad[1:], t_vmax, hol_b, close_centr_msk[1:])
        v_ang_norm = _stat_holland_2010(d_centr[1:], t_vmax, t_rad[1:], hol_b,
                                        close_centr_msk[1:], hol_x)
    else:
        raise NotImplementedError

    # vectorial angular velocity
    hemisphere = 'N'
    if np.count_nonzero(t_lat < 0) > np.count_nonzero(t_lat > 0):
        hemisphere = 'S'
    v_ang_rotate = [1.0, -1.0] if hemisphere == 'N' else [-1.0, 1.0]
    v_ang_dir = np.array(v_ang_rotate)[..., :] * v_centr_normed[1:, :, ::-1]
    v_ang = np.zeros_like(v_ang_dir)
    v_ang[close_centr_msk[1:]] = v_ang_norm[close_centr_msk[1:], None] \
                                 * v_ang_dir[close_centr_msk[1:]]

    # Influence of translational speed decreases with distance from eye.
    # The "absorbing factor" is according to the following paper (see Fig. 7):
    #
    #   Mouton, F., & Nordbeck, O. (1999). Cyclone Database Manager. A tool
    #   for converting point data from cyclone observations into tracks and
    #   wind speed profiles in a GIS. UNED/GRID-Geneva.
    #   https://unepgrid.ch/en/resource/19B7D302
    #
    t_rad_bc = np.broadcast_arrays(t_rad[:, None], d_centr)[0]
    v_trans_corr = np.zeros_like(d_centr)
    v_trans_corr[close_centr_msk] = np.fmin(
        1, t_rad_bc[close_centr_msk] / d_centr[close_centr_msk])

    # add angular and corrected translational velocity vectors
    v_full = v_trans[1][1:, None, :] * v_trans_corr[1:, :, None] + v_ang
    v_full[np.isnan(v_full)] = 0

    windfields = np.zeros((npositions, nreachable, 2), dtype=np.float64)
    windfields[1:, :, :] = v_full
    [reachable_centr_idx] = track_centr_msk.nonzero()
    return windfields, reachable_centr_idx

def _close_centroids(t_lat, t_lon, centroids, buffer=CENTR_NODE_MAX_DIST_DEG):
    """Check whether centroids lay within a rectangular buffer around track positions

    The longitudinal coordinates are assumed to be normalized around a central longitude. This
    makes sure that the buffered bounding box around the track doesn't cross the antimeridian.

    The only hypothetical problem occurs when a TC track is travelling more than 349 degrees in
    longitude because that's when adding a buffer of 5.5 degrees might cross the antimeridian.
    Of course, this case is physically impossible.

    Parameters
    ----------
    t_lat : np.array of shape (npositions,)
        Latitudinal coordinates of track positions.
    t_lon : np.array of shape (npositions,)
        Longitudinal coordinates of track positions, normalized around a central longitude.
    centroids : np.array of shape (ncentroids, 2)
        Coordinates of centroids, each row is a pair [lat, lon].
    buffer : float (optional)
        Size of the buffer. Default: CENTR_NODE_MAX_DIST_DEG.

    Returns
    -------
    mask : np.array of shape (ncentroids,)
        Mask that is True for close centroids and False for other centroids.
    """
    centr_lat, centr_lon = centroids[:, 0], centroids[:, 1]
    # check for each track position which centroids are within buffer, uses NumPy's broadcasting
    mask = ((t_lat[:, None] - buffer < centr_lat[None])
            & (centr_lat[None] < t_lat[:, None] + buffer)
            & (t_lon[:, None] - buffer < centr_lon[None])
            & (centr_lon[None] < t_lon[:, None] + buffer))
    # for each centroid, check whether it is in the buffer for any of the track positions
    return mask.any(axis=0)

def _vtrans(t_lat, t_lon, t_tstep, metric="equirect"):
    """Translational vector and velocity at each track node.

    Parameters
    ----------
    t_lat : np.array
        track latitudes
    t_lon : np.array
        track longitudes
    t_tstep : np.array
        track time steps
    metric : str, optional
        Specify an approximation method to use for earth distances: "equirect" (faster) or
        "geosphere" (more accurate). See `dist_approx` function in `climada.util.coordinates`.
        Default: "equirect".

    Returns
    -------
    v_trans_norm : np.array
        Same shape as input, the first velocity is always 0.
    v_trans : np.array
        Directional vectors of velocity.
    """
    v_trans = np.zeros((t_lat.size, 2))
    v_trans_norm = np.zeros((t_lat.size,))
    norm, vec = u_coord.dist_approx(t_lat[:-1, None], t_lon[:-1, None],
                                    t_lat[1:, None], t_lon[1:, None],
                                    log=True, normalize=False, method=metric)
    v_trans[1:, :] = vec[:, 0, 0]
    v_trans[1:, :] *= KMH_TO_MS / t_tstep[1:, None]
    v_trans_norm[1:] = norm[:, 0, 0]
    v_trans_norm[1:] *= KMH_TO_MS / t_tstep[1:]

    # limit to 30 nautical miles per hour
    msk = (v_trans_norm > 30 * KN_TO_MS)
    fact = 30 * KN_TO_MS / v_trans_norm[msk]
    v_trans[msk, :] *= fact[:, None]
    v_trans_norm[msk] *= fact
    return v_trans_norm, v_trans

def _bs_holland_2008(v_trans, penv, pcen, prepcen, lat, tint):
    """Holland's 2008 b-value estimate for sustained surface winds.

    Unlike the original 1980 formula (see `_B_holland_1980`), this approach does not require any
    wind speed measurements, but is based on the more reliable pressure information.

    The parameter applies to 1-minute sustained winds at 10 meters above ground.
    It is taken from equation (11) in the following paper:

    Holland, G. (2008). A revised hurricane pressure-wind model. Monthly
    Weather Review, 136(9), 3432–3445. https://doi.org/10.1175/2008MWR2395.1

    For reference, it reads

    b_s = -4.4 * 1e-5 * (penv - pcen)^2 + 0.01 * (penv - pcen)
          + 0.03 * (dp/dt) - 0.014 * |lat| + 0.15 * (v_trans)^hol_xx + 1.0

    where `dp/dt` is the time derivative of central pressure and `hol_xx` is Holland's x
    parameter: hol_xx = 0.6 * (1 - (penv - pcen) / 215)

    The equation for b_s has been fitted statistically using hurricane best track records for
    central pressure and maximum wind. It therefore performs best in the North Atlantic.

    Furthermore, b_s has been fitted under the assumption of a "cyclostrophic" wind field which
    means that the influence from Coriolis forces is assumed to be small. This is reasonable close
    to the radius of maximum wind where the Coriolis term (r*f/2) is small compared to the rest
    (see `_stat_holland`). More precisely: At the radius of maximum wind speeds, the typical order
    of the Coriolis term is 1 while wind speed is 50 (which changes away from the
    radius of maximum winds and as the TC moves away from the equator).

    Parameters
    ----------
    v_trans : np.array
        Translational wind (m/s).
    penv : np.array
        Environmental pressure (hPa).
    pcen : np.array
        Central pressure (hPa).
    prepcen : np.array
        Central pressure (hPa) at previous track position.
    lat : np.array
        Latitude (degrees).
    tint : np.array
        Time step (h).

    Returns
    -------
    b_s : np.array
        Holland b-value
    """
    pdelta = penv - pcen
    hol_xx = 0.6 * (1. - pdelta / 215)
    hol_b = -4.4e-5 * pdelta**2 + 0.01 * pdelta + \
        0.03 * (pcen - prepcen) / tint - 0.014 * abs(lat) + \
        0.15 * v_trans**hol_xx + 1.0
    return np.clip(hol_b, 1, 2.5)

def _v_max_s_holland_2008(penv, pcen, b_s):
    """Compute maximum surface winds from pressure according to Holland 2008.

    This function implements equation (11) in the following paper:

    Holland, G. (2008). A revised hurricane pressure-wind model. Monthly
    Weather Review, 136(9), 3432–3445. https://doi.org/10.1175/2008MWR2395.1

    For reference, it reads

    v_ms = [b_s / (rho * e) * (penv - pcen)]^0.5

    where `b_s` is Holland b-value (see `_bs_holland_2008`), e is Euler's number, rho is the
    density of air, `penv` is environmental, and `pcen` is central pressure.

    Parameters
    ----------
    penv : np.array
        Environmental pressure (hPa).
    pcen : np.array
        Central pressure (hPa).
    b_s : np.array
        Holland's b-parameter according to `_bs_holland_2008`.

    Returns
    -------
    v_max_s : np.array
        Maximum surface winds.
    """
    # the factor 100 is from conversion between mbar and pascal
    v_squared = b_s / (RHO_AIR * np.exp(1)) * 100 * (penv - pcen)
    return np.sqrt(v_squared)

def _B_holland_1980(gradient_winds, penv, pcen):
    """Holland's 1980 B-value computation for gradient-level winds.

    The parameter applies to gradient-level winds (about 1000 metres above the earth's surface).
    The formula for B is derived from equations (5) and (6) in the following paper:

    Holland, G.J. (1980): An Analytic Model of the Wind and Pressure Profiles
    in Hurricanes. Monthly Weather Review 108(8): 1212–1218.
    https://doi.org/10.1175/1520-0493(1980)108<1212:AAMOTW>2.0.CO;2

    For reference, inserting (6) into (5) and solving for B at r = RMW yields:

    B = v^2 * e * rho / (penv - pcen)

    where v are maximum gradient-level winds `gradient_winds`, e is Euler's number, rho is the
    density of air, `penv` is environmental, and `pcen` is central pressure.

    Parameters
    ----------
    gradient_winds : np.array
        Maximum gradient-level wind speeds (m/s) of the tropical cyclone. If your data are maximum
        surface wind speeds (e.g. from IBTrACS), make sure to subtract translational wind speed and
        convert to gradient-level winds first.
    penv : np.array
        Environmental pressure (hPa).
    pcen : np.array
        Central pressure (hPa).

    Returns
    -------
    B : np.array
        Holland b-value
    """
    # the factor 100 is from conversion between mbar and pascal
    pdelta = 100 * (penv - pcen)
    hol_b = gradient_winds**2 * np.exp(1) * RHO_AIR / np.fmax(np.spacing(1), pdelta)
    return np.clip(hol_b, 1, 2.5)

def _x_holland_2010(d_centr, r_max, v_max_s, hol_b, close_centr, v_n=17.0, r_n=300e3):
    """Compute exponent for wind model according to Holland et al. 2010.

    This function implements equation (10) from the following paper:

    Holland et al. (2010): A Revised Model for Radial Profiles of Hurricane Winds. Monthly
    Weather Review 138(12): 4393–4401. https://doi.org/10.1175/2010MWR3317.1

    For reference, it reads

    x = 0.5  [for r < r_max]
    x = 0.5 + (r - r_max) * (x_n - 0.5) / (r_n - r_max)  [for r >= r_max]

    The peripheral exponent x_n is adjusted to fit the peripheral observation of wind speeds `v_n`
    at radius `r_n`.

    Parameters
    ----------
    d_centr : np.array of shape (nnodes, ncentroids)
        Distance (m) between centroids and track nodes.
    r_max : np.array of shape (nnodes,)
        Radius (m) of maximum winds at each track node.
    v_max_s : np.array of shape (nnodes,)
        Maximum surface winds (m/s) at each track node.
    hol_b : np.array of shape (nnodes,)
        Holland's b parameter at each track node.
    close_centr : np.array of shape (nnodes, ncentroids)
        Mask indicating for each track node which centroids are within reach of the windfield.
    v_n : np.array of shape (nnodes,) or float, optional
        Peripheral wind speeds (m/s) at radius `r_n` outside of radius of maximum winds `r_max`.
        In absence of a second wind speed measurement, this value defaults to 17 m/s following
        Holland et al. 2010 (at a radius of 300 km).
    r_n : np.array of shape (nnodes,) or float, optional
        Radius (m) where the peripheral wind speed `v_n` is measured (or assumed).
        In absence of a second wind speed measurement, this value defaults to 300 km following
        Holland et al. 2010.

    Returns
    -------
    x : np.array of shape (nnodes, ncentroids)
        Exponents according to Holland et al. 2010.
    """
    x = np.zeros_like(d_centr)
    r_max, v_max_s, hol_b, d_centr, v_n, r_n = [
        ar[close_centr] for ar in np.broadcast_arrays(
            r_max[:, None], v_max_s[:, None], hol_b[:, None], d_centr,
            np.atleast_1d(v_n)[:, None], np.atleast_1d(r_n)[:, None])
    ]

    # compute peripheral exponent from second measurement
    r_max_norm = (r_max / r_n)**hol_b
    x_n = np.log(v_n / v_max_s) / np.log(r_max_norm * np.exp(1 - r_max_norm))

    x[close_centr] = 0.5 + np.fmax(0, d_centr - r_max) * (x_n - 0.5) / (r_n - r_max)
    x[close_centr] = np.clip(x[close_centr], 0.0, 0.5)
    return x


def _stat_holland_2010(d_centr, v_max_s, r_max, hol_b, close_centr, x):
    """Symmetric and static surface wind fields (in m/s) according to Holland et al. 2010

    This function applies the cyclostrophic surface wind model expressed in equation (6) from

    Holland et al. (2010): A Revised Model for Radial Profiles of Hurricane Winds. Monthly
    Weather Review 138(12): 4393–4401. https://doi.org/10.1175/2010MWR3317.1

    More precisely, this function implements the following equation:

    V(r) = v_max_s * [(r_max / r)^b_s * e^(1 - (r_max / r)^b_s)]^x

    In terms of this function's arguments, b_s is `hol_b` and r is `d_centr`.

    Parameters
    ----------
    d_centr : np.array of shape (nnodes, ncentroids)
        Distance (m) between centroids and track nodes.
    v_max_s : np.array of shape (nnodes,)
        Maximum surface wind speeds (m/s) of the tropical cyclone according to
        `_v_max_s_holland_2008`.
    r_max : np.array of shape (nnodes,)
        Radius (m) of maximum winds at each track node.
    hol_b : np.array of shape (nnodes,)
        Holland's b parameter at each track node according to `_bs_holland_2008`.
    close_centr : np.array of shape (nnodes, ncentroids)
        Mask indicating for each track node which centroids are within reach of the windfield.
    x : np.array of shape (nnodes, ncentroids) or float, optional
        The exponent according to `_x_holland_2010`.

    Returns
    -------
    v_ang : np.array (nnodes, ncentroids)
        Absolute values of wind speeds (m/s) in angular direction.
    """
    v_ang = np.zeros_like(d_centr)
    d_centr, v_max_s, r_max, hol_b, x = [
        ar[close_centr] for ar in np.broadcast_arrays(
            d_centr, v_max_s[:, None], r_max[:, None], hol_b[:, None], x)
    ]

    r_max_norm = (r_max / d_centr)**hol_b
    v_ang[close_centr] = v_max_s * (r_max_norm * np.exp(1 - r_max_norm))**x
    return v_ang

def _stat_holland_1980(d_centr, r_max, hol_b, penv, pcen, lat, close_centr, cyclostrophic=False):
    """Symmetric and static wind fields (in m/s) according to Holland 1980.

    This function applies the gradient wind model expressed in equation (4) (combined with
    equation (6)) from

    Holland, G.J. (1980): An Analytic Model of the Wind and Pressure Profiles in Hurricanes.
    Monthly Weather Review 108(8): 1212–1218.

    More precisely, this function implements the following equation:

    V(r) = [(B/rho) * (r_max/r)^B * (penv - pcen) * e^(-(r_max/r)^B) + (r*f/2)^2]^0.5 - (r*f/2)

    In terms of this function's arguments, B is `hol_b` and r is `d_centr`.
    The air density rho is assumed to be constant while the Coriolis parameter f is computed
    from the latitude `lat` using the constant rotation rate of the earth.

    Even though the equation has been derived originally for gradient winds (when combined with the
    output of `_B_holland_1980`), it can be used for surface winds by adjusting the parameter
    `hol_b` (see function `_bs_holland_2008`).

    Parameters
    ----------
    d_centr : np.array of shape (nnodes, ncentroids)
        Distance (m) between centroids and track nodes.
    r_max : np.array of shape (nnodes,)
        Radius (m) of maximum winds at each track node.
    hol_b : np.array of shape (nnodes,)
        Holland's b parameter at each track node.
    penv : np.array of shape (nnodes,)
        Environmental pressure (hPa) at each track node.
    pcen : np.array of shape (nnodes,)
        Central pressure (hPa) at each track node.
    lat : np.array of shape (nnodes,)
        Latitudinal coordinate of each track node.
    close_centr : np.array of shape (nnodes, ncentroids)
        Mask indicating for each track node which centroids are within reach of the windfield.
    cyclostrophic : bool, optional
        If True, don't apply the influence of the Coriolis force (set the Coriolis terms to 0).
        Default: False

    Returns
    -------
    v_ang : np.array (nnodes, ncentroids)
        Absolute values of wind speeds (m/s) in angular direction.
    """
    v_ang = np.zeros_like(d_centr)
    d_centr, r_max, hol_b, penv, pcen, lat = [
        ar[close_centr] for ar in np.broadcast_arrays(
            d_centr, r_max[:, None], hol_b[:, None], penv[:, None], pcen[:, None], lat[:, None])
    ]

    r_coriolis = 0
    if not cyclostrophic:
        # Coriolis parameter with earth rotation rate 7.29e-5
        f_coriolis = 2 * 0.0000729 * np.sin(np.radians(np.abs(lat)))

        # d_centr is in km, convert to m (factor 1000) and apply Coriolis parameter
        r_coriolis = 0.5 * 1000 * d_centr * f_coriolis

    # the factor 100 is from conversion between mbar and pascal
    r_max_norm = (r_max / d_centr)**hol_b
    sqrt_term = 100 * hol_b / RHO_AIR * r_max_norm * (penv - pcen) \
                * np.exp(-r_max_norm) + r_coriolis**2

    v_ang[close_centr] = np.sqrt(np.fmax(0, sqrt_term)) - r_coriolis
    return v_ang



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define functions to handle dates and times in climada
"""
import logging
import datetime as dt
import numpy as np
import pandas as pd

LOGGER = logging.getLogger(__name__)

def date_to_str(date):
    """Compute date string in ISO format from input datetime ordinal int.
    Parameters
    ----------
    date : int or list or np.array
        input datetime ordinal

    Returns
    -------
    str or list(str)
    """
    try:
        date_int = int(date)
        return dt.date.fromordinal(date_int).isoformat()
    except TypeError:
        return [dt.date.fromordinal(i_date).isoformat() for i_date in date]


def str_to_date(date):
    """Compute datetime ordinal int from input date string in ISO format.
    Parameters
    ----------
    date : str or list
        idate string in ISO format, e.g. '2018-04-06'

    Returns
    -------
    int
    """
    if isinstance(date, str):
        year, mounth, day = (int(val) for val in date.split('-'))
        return dt.date(year, mounth, day).toordinal()

    all_date = []
    for i_date in date:
        year, mounth, day = (int(val) for val in i_date.split('-'))
        all_date.append(dt.date(year, mounth, day).toordinal())
    return all_date

def datetime64_to_ordinal(datetime):
    """Converts from a numpy datetime64 object to an ordinal date.
    See https://stackoverflow.com/a/21916253 for the horrible details.
    Parameters
    ----------
    datetime : np.datetime64, or list or np.array
        date and time

    Returns
    -------
    int
    """
    if isinstance(datetime, np.datetime64):
        return pd.to_datetime(datetime.tolist()).toordinal()

    return [pd.to_datetime(i_dt.tolist()).toordinal() for i_dt in datetime]

def last_year(ordinal_vector):
    """Extract first year from ordinal date

    Parameters
    ----------
    ordinal_vector : list or np.array
        input datetime ordinal

    Returns
    -------
    int
    """
    return dt.date.fromordinal(np.max(ordinal_vector)).year

def first_year(ordinal_vector):
    """Extract first year from ordinal date

    Parameters
    ----------
    ordinal_vector : list or np.array
        input datetime ordinal

    Returns
    -------
    int
    """
    return dt.date.fromordinal(np.min(ordinal_vector)).year



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test plot module.
"""

import unittest
import cartopy
import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs

import climada.util.plot as u_plot

class TestFuncs(unittest.TestCase):

    def test_get_transform_4326_pass(self):
        """Check _get_transformation for 4326 epsg."""
        res, unit = u_plot.get_transformation('epsg:4326')
        self.assertIsInstance(res, cartopy.crs.PlateCarree)
        self.assertEqual(unit, '°')

    def test_get_transform_3395_pass(self):
        """Check that assigned attribute is correctly set."""
        res, unit = u_plot.get_transformation('epsg:3395')
        self.assertIsInstance(res, cartopy.crs.Mercator)
        self.assertEqual(unit, 'm')

    def test_get_transform_3035_pass(self):
        """Check that assigned attribute is correctly set."""
        res, unit = u_plot.get_transformation('epsg:3035')
        self.assertIsInstance(res, cartopy.crs.Projection)
        self.assertEqual(res.epsg_code, 3035)
        self.assertEqual(unit, 'm')

class TestPlots(unittest.TestCase):

    def test_geo_scatter_categorical(self):
        """Plots ones with geo_scatteR_categorical"""
        # test default with one plot
        values = np.array([1, 2.0, 1, 'a'])
        coord = np.array([[26, 0], [26, 1], [28, 0], [29, 1]])
        u_plot.geo_scatter_categorical(values, coord, 'value', 'test plot',
                        pop_name=True)
        plt.close()

        #test multiple plots with non default kwargs
        values = np.array([[1, 2.0, 1, 'a'], [0, 0, 0, 0]])
        coord = np.array([[26, 0], [26, 1], [28, 0], [29, 1]])
        u_plot.geo_scatter_categorical(values, coord, 'value', 'test plot',
                        cat_name={0: 'zero',
                                  1: 'int',
                                  2.0: 'float',
                                  'a': 'string'},
                        pop_name=False, cmap='Set1')
        plt.close()

        #test colormap warning
        values = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11],
                           [12, 13, 14, 15]])
        coord = np.array([[26, 0], [26, 4], [28, 0], [29, 1]])
        u_plot.geo_scatter_categorical(values, coord, 'value', 'test plot',
                        pop_name=False, cmap='Set1')

        plt.close()

        #test colormap warning with 256 colors
        values = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11],
                           [12, 13, 14, 15]])
        coord = np.array([[26, 0], [26, 1], [28, 0], [29, 1]])
        u_plot.geo_scatter_categorical(values, coord, 'value', 'test plot',
                        pop_name=False, cmap='tab20c')
        plt.close()

    def test_geo_scatter_from_array(self):
        values = np.array([1, 2.0, 1, 1])
        coord = np.array([[-17, 178], [-10, 180], [-27, 175], [-16, 186]])
        var_name = 'test'
        title = 'test'
        projection = ccrs.PlateCarree()
        cmap = 'viridis'
        ax = u_plot.geo_scatter_from_array(values, coord, var_name, title,
                               pop_name=True, extend='neither',
                               shapes=True, axes=None, proj=projection,
                               figsize=(9, 13), cmap=cmap)
        self.assertEqual(var_name, ax.get_title())
        self.assertAlmostEqual(np.max(values), ax.collections[0].colorbar.vmax)
        self.assertAlmostEqual(np.min(values), ax.collections[0].colorbar.vmin)
        self.assertEqual(cmap, ax.collections[0].cmap.name)
        plt.close()

    def test_geo_bin_from_array(self):
        values = np.array([1, 2.0, 5, 1])
        coord = np.array([[-10, 17], [-30, 20], [5, 75], [-16, 20]])
        var_name = 'test'
        title = 'test'
        projection = ccrs.PlateCarree()
        cmap = 'viridis'
        ax = u_plot.geo_bin_from_array(values, coord, var_name, title,
                                           pop_name=True, extend='neither',
                                           shapes=True, axes=None, proj=projection,
                                           figsize=(9, 13), cmap=cmap)
        self.assertEqual(var_name, ax.get_title())
        self.assertAlmostEqual(np.max(values), ax.collections[0].colorbar.vmax)
        self.assertAlmostEqual(np.min(values), ax.collections[0].colorbar.vmin)
        self.assertEqual(cmap, ax.collections[0].cmap.name)
        plt.close()

    def test_geo_im_from_array(self):
        values = np.array([1, 2.0, 5, 1])
        coord = np.array([[-17, 178], [-10, 180], [-27, 175], [-16, 186]])
        var_name = 'test'
        title = 'test'
        projection = ccrs.PlateCarree()
        cmap = 'viridis'
        ax = u_plot.geo_im_from_array(values, coord, var_name, title,
                      proj=projection, smooth=True, axes=None, figsize=(9, 13), cmap=cmap)
        self.assertEqual(var_name, ax.get_title())
        self.assertAlmostEqual(np.max(values), ax.collections[0].colorbar.vmax)
        self.assertAlmostEqual(np.min(values), ax.collections[0].colorbar.vmin)
        self.assertEqual(cmap, ax.collections[0].cmap.name)
        plt.close()


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestFuncs)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPlots))
    unittest.TextTestRunner(verbosity=2).run(TESTS)
